#include <QFile>
#include <iostream>

#include "tradingModel.h"
#include "simpleMovingAverage.h"
#include "priceDataSet.h"
#include "stockPortfolio.h"
#include "tradingUtils.h"
#include "modelParameters.h"

// Trading model class that executes trading logic for a crossover based strategy

tradingModel::tradingModel()
{
}

void tradingModel::test()
{
    priceData_->test();
}

// Initialize the model
bool tradingModel::init(std::shared_ptr<modelParameters> inParams)
{
    // Start with a portfolio of $1,000
    double portfolioStartValue = 1000;
    modelPortfolio_ = std::unique_ptr<stockPortfolio>(new stockPortfolio(portfolioStartValue));

    // We start out as sold
    setTradeState(SOLD);

    // Grab input parameters and store them to local var modelParams_
    modelParams_ = inParams;

    // File name -> price data load
    bool priceDataLoaded = loadPriceData();

    // Moving average lengths
    movingAvgLen1_ = modelParams_->getMovingAvgLen1();
    movingAvgLen2_ = modelParams_->getMovingAvgLen2();

    bool movingAverageLengthsValid = ((movingAvgLen1_ > 0) && (movingAvgLen2_ > 0));

    return (priceDataLoaded && movingAverageLengthsValid);
}


// Read the data file and load it into a price data object
bool tradingModel::loadPriceData()
{
    bool fileReadSuccess = false;

    // Get the data filename
    QString priceDataFilename = modelParams_->getDataFilename();

    // Extra check to make sure we've received a file name
    if (priceDataFilename.isEmpty())
        return false;

    // Create new price data object
    priceData_ = std::unique_ptr<priceDataSet>(new priceDataSet());

    // Set up a QFile for reading
    QFile file(priceDataFilename);
    // Check if file exists
    if (file.exists())
    {
        // Check if the file is readable
        if ( !file.open(QFile::ReadOnly | QFile::Text) )
        {
            std::cerr << "Problem opening file " << priceDataFilename.toStdString() << std::endl;
            return false;
        }
        fileReadSuccess = priceData_->Read(&file);
    }
    return fileReadSuccess;
}


// Functions to manage our model's trade state (eg., state machine)
void tradingModel::setTradeState(TRADE_STATE inState)
{
    tradeState_ = inState;
}

TRADE_STATE tradingModel::getTradeState()
{
    return tradeState_;
}

// Run the model
void tradingModel::run()
{

    // Set up our moving averages
    std::unique_ptr<simpleMovingAverage> mAvg1 = std::unique_ptr<simpleMovingAverage>(new simpleMovingAverage(movingAvgLen1_));
    std::unique_ptr<simpleMovingAverage> mAvg2 = std::unique_ptr<simpleMovingAverage>(new simpleMovingAverage(movingAvgLen2_));

    double open,high,low,close;
    unsigned today;

    double mAvg1Val = 0.0;
    double mAvg2Val = 0.0;

    // MAIN TRADING ITERATOR
    for(unsigned long todayIndex = 0; todayIndex < priceData_->getNumRows(); todayIndex++)
    {
        // Grab price data into local variables
        today = priceData_->getDate(todayIndex);
        open = priceData_->getOpen(todayIndex);
        high = priceData_->getHigh(todayIndex);
        low = priceData_->getLow(todayIndex);
        close = priceData_->getClose(todayIndex);

        // PHASE 1 - Decisions ------------------------------------------------

        // Buy signal
        if ((mAvg1Val > mAvg2Val) && (getTradeState() == SOLD))
        {
            setTradeState(BUY);
        }
        // Sell signal
        else if ((mAvg1Val < mAvg2Val) && (getTradeState() == BOUGHT))
        {
            setTradeState(SELL);
        }

        // PHASE 2 - Actions --------------------------------------------------

        // Execute buy signal
        if (getTradeState() == BUY)
        {
            std::cout << "BUY - ";
            std::cout << today << " ";
            double cashAvailable = modelPortfolio_->getCashValue();
            unsigned numSharesToBuy = static_cast<unsigned>(cashAvailable/close);
            std:: cout << numSharesToBuy << " shares at " << close << std::endl;

            modelPortfolio_->doBuy(close,numSharesToBuy);

            setTradeState(BOUGHT);
        }
        // Execute sell signal
        if (getTradeState() == SELL)
        {
            std::cout << "SELL - ";
            std::cout << today << " ";
            unsigned numSharesSold = modelPortfolio_->getNumShares();
            std:: cout << numSharesSold << " shares at " << close << std::endl;

            modelPortfolio_->doSell(close);

            setTradeState(SOLD);
        }

        // PHASE 3 - Updates / Calculations -----------------------------------

        // We're trading based on what happened "yesterday"
        // Update moving averages
        mAvg1->addDataPoint(close);
        mAvg2->addDataPoint(close);
        mAvg1Val = tradingUtils::fix2D(mAvg1->getCurrentValue());
        mAvg2Val = tradingUtils::fix2D(mAvg2->getCurrentValue());
        // Update portfolio value
        modelPortfolio_->update(close);

    }
}

// Print out various stats related to the most recent run
void tradingModel::printStats()
{
    std::cout << std::endl;

    // Portfolio start / end and percent
    std::cout << "Portfolio start = " << modelPortfolio_->getStartValue() << std::endl;
    std::cout << "Portfolio end = " << modelPortfolio_->getTotalValue() << std::endl;
    double portfolioTotalPercentChange =
            tradingUtils::fix2D((modelPortfolio_->getTotalValue()/modelPortfolio_->getStartValue())-1);
    std::cout << "  pct change = " << portfolioTotalPercentChange << std::endl;

    // Portfolio max value and percent
    std::cout << "Portfolio max = " << modelPortfolio_->getMaxValue() << std::endl;
    double portfolioMaxPercentChange =
            tradingUtils::fix2D((modelPortfolio_->getMaxValue()/modelPortfolio_->getStartValue())-1);
    std::cout << "  pct change = " << portfolioMaxPercentChange << std::endl;

    // Portfolio min value and percent
    std::cout << "Portfolio min = " << modelPortfolio_->getMinValue() << std::endl;
    double portfolioMinPercentChange =
            tradingUtils::fix2D((modelPortfolio_->getMinValue()/modelPortfolio_->getStartValue())-1);
    std::cout << "  pct change = " << portfolioMinPercentChange << std::endl;

    // Trades both positive and negative
    std::cout << "Num sell trades = " << modelPortfolio_->getNumSellTrades() << std::endl;
    std::cout << "  Num positive trades = " << modelPortfolio_->getNumTradesPos();
    double posTradePercent =
            tradingUtils::fix2D(modelPortfolio_->getNumTradesPos()/static_cast<double>(modelPortfolio_->getNumSellTrades()));
    std::cout << " (" << posTradePercent << " pct)" << std::endl;
    std::cout << "  Num negative trades = " << modelPortfolio_->getNumTradesNeg();
    double negTradePercent =
            tradingUtils::fix2D(modelPortfolio_->getNumTradesNeg()/static_cast<double>(modelPortfolio_->getNumSellTrades()));
    std::cout << " (" << negTradePercent << " pct)" << std::endl;
}
